@@ -3,7 +3,7 @@
 // Purpose:     FFMPEG Media Decoder
 // Author:      Alex Thuering
 // Created:     21.07.2007
-// RCS-ID:      $Id: mediadec_ffmpeg.cpp,v 1.27 2014/05/15 19:52:58 ntalex Exp $
+// RCS-ID:      $Id: mediadec_ffmpeg.cpp,v 1.31 2014/12/30 12:17:08 ntalex Exp $
 // Copyright:   (c) Alex Thuering
 // Licence:     wxWindows licence
 /////////////////////////////////////////////////////////////////////////////
@@ -26,6 +26,11 @@ extern "C" {
 #include <libavutil/mathematics.h>
 }
 
+#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(55, 28, 1)
+#define av_frame_alloc avcodec_alloc_frame
+#define av_frame_free avcodec_free_frame
+#endif
+
 wxFfmpegMediaDecoder::wxFfmpegMediaDecoder(): m_formatCtx(NULL), m_videoStream(-1), m_codecCtx(NULL), m_frame(NULL),
 		m_width(0), m_height(0) {
 	// nothing to do
@@ -116,11 +121,11 @@ float wxFfmpegMediaDecoder::GetFps() {
 	float result = -1;
 	for (unsigned int i = 0; i < m_formatCtx->nb_streams; i++) {
 		AVStream* st = m_formatCtx->streams[i];
-		if (st->codec->codec_type == AVMEDIA_TYPE_VIDEO && st->r_frame_rate.num && st->r_frame_rate.den) {
-			if (st->r_frame_rate.num > st->r_frame_rate.den)
-				result = ((float) st->r_frame_rate.num) / st->r_frame_rate.den;
+		if (st->codec->codec_type == AVMEDIA_TYPE_VIDEO && st->avg_frame_rate.num && st->avg_frame_rate.den) {
+			if (st->avg_frame_rate.num > st->avg_frame_rate.den)
+				result = ((float) st->avg_frame_rate.num) / st->avg_frame_rate.den;
 			else
-				result = ((float) st->r_frame_rate.den) / st->r_frame_rate.num;
+				result = ((float) st->avg_frame_rate.den) / st->avg_frame_rate.num;
 			break;
 		}
 	}
@@ -216,11 +221,7 @@ bool wxFfmpegMediaDecoder::BeginDecode(int width, int height)
         m_height = w < width ? height : h;
     }
     // allocate video frame
-#if LIBAVCODEC_VERSION_MAJOR >= 55
     m_frame = av_frame_alloc();
-#else
-    m_frame = avcodec_alloc_frame();
-#endif
     if (!m_frame) {
         avcodec_close(m_codecCtx);
         m_codecCtx = NULL;
@@ -232,11 +233,16 @@ bool wxFfmpegMediaDecoder::BeginDecode(int width, int height)
 bool wxFfmpegMediaDecoder::SetPosition(double pos, bool keyFrame) {
     if (m_formatCtx == NULL)
         return false;
+	if (!m_frame && !BeginDecode())
+		return false;
     int64_t timestamp = (int64_t) (pos * AV_TIME_BASE);
     if (m_formatCtx->start_time != (int64_t)AV_NOPTS_VALUE)
         timestamp += m_formatCtx->start_time;
-    return av_seek_frame(m_formatCtx, -1, timestamp,
+    avcodec_flush_buffers(m_codecCtx);
+    bool res = av_seek_frame(m_formatCtx, -1, timestamp,
     		keyFrame ? AVSEEK_FLAG_BACKWARD : AVSEEK_FLAG_ANY|AVSEEK_FLAG_BACKWARD) >= 0;
+    avcodec_flush_buffers(m_codecCtx);
+    return res;
 }
 
 double wxFfmpegMediaDecoder::GetPosition() {
@@ -266,14 +272,17 @@ wxImage wxFfmpegMediaDecoder::GetNextFrame() {
 			if (frameFinished) {
 				SwsContext* imgConvertCtx = sws_getContext(m_codecCtx->width, m_codecCtx->height, m_codecCtx->pix_fmt,
 						m_width, m_height, PIX_FMT_RGB24, SWS_BICUBIC, NULL, NULL, NULL);
-				if (imgConvertCtx == NULL)
+				if (imgConvertCtx == NULL) {
+					av_free_packet(&packet);
 					return wxImage();
+				}
 
 				wxImage img(m_width, m_height);
 				uint8_t *rgbSrc[3] = { img.GetData(), NULL, NULL };
 				int rgbStride[3] = { 3 * m_width, 0, 0 };
 				sws_scale(imgConvertCtx, m_frame->data, m_frame->linesize, 0, m_codecCtx->height, rgbSrc, rgbStride);
 				av_free_packet(&packet);
+				sws_freeContext(imgConvertCtx);
 				return img;
 			}
 		}
@@ -283,10 +292,7 @@ wxImage wxFfmpegMediaDecoder::GetNextFrame() {
 	return wxImage();
 }
 
-void wxFfmpegMediaDecoder::EndDecode()
-{
-    if (m_frame)
-        av_free(m_frame);
-    m_frame = NULL;
+void wxFfmpegMediaDecoder::EndDecode() {
+	av_frame_free(&m_frame);
     CloseVideoDecoder();
 }
